#ifndef FSGUICANVAS_IS_INCLUDED
#define FSGUICANVAS_IS_INCLUDED
/* { */

#include "fsguibase.h"

////////////////////////////////////////////////////////////

/*! This class makes a GUI canvas.  GUI canvas can have up to a menu and a main dialog.

	A GUI canvas routes events coming from the operating system.

    If a modal-dialog chain is attached to this canvas, all events are directed to 
    the bottom-most dialog of the modal-dialog chain of the canvas.

	Else, if a modal-dialog chain is attached to the main menu, all events are directed to 
	the bottom-most dialog of the modal-dialog chain of the main menu.

	Else, if a modal-dialog chain is attached to the main dialog, all events are directed to 
	the bottom-most dialog of the modal-dialog chain of the main dialog.

    The modal dialog is responsible for closing it by CloseModalDialog function.
  */
class FsGuiCanvas : public FsGuiControlBase
{
private:
	class FsGuiPopUpMenu *mainMenu;
	class FsGuiStatusBar *topStatusBar,*bottomStatusBar;
	YsArray <class FsGuiDialog *> dlgArray;
	class FsGuiDialog *activeDlg;

	int prevMx,prevMy;  // Cache previous mouse cursor position
	YSBOOL prevLb,prevMb,prevRb; // Cache previous mouse button state
	int prevWid,prevHei; // Cache previous window size
	YSBOOL needRedraw;

public:
	/*! Default constructor */
	FsGuiCanvas();

	/*! Initializes the object */
	void Initialize(void);

	/*! This function sets a pointer to the main menu. */
	void SetMainMenu(class FsGuiPopUpMenu *mainMenu);

	/*! This function sets a status bar at the top of the window. */
	void SetTopStatusBar(class FsGuiStatusBar *topBar);

	/*! This function returns a pointer to the top status bar.  NULL will be returned if no top status bar is set. */
	class FsGuiStatusBar *GetTopStatusBar(void) const;

	/*! This function sets a status bar at the top of the window. */
	void SetBottomStatusBar(class FsGuiStatusBar *bottomBar);

	/*! This function returns a pointer to the bottom status bar.    NULL will be returned if no top status bar is set. */
	class FsGuiStatusBar *GetBottomStatusBar(void) const;

	/*! This function returns a pointer to the main menu.  If no main menu is 
	    set, NULL will be returned. */
	class FsGuiPopUpMenu *GetMainMenu(void) const;

	/*! This function adds a pointer to a modeless dialog. 
	    Returns YSERR if the incoming dialog is already in the dialog array. */
	YSRESULT AddDialog(class FsGuiDialog *dlg);

	/*! This function removes a pointer to a modeless dialog. 
	    Returns YSERR if the dialog is not in the dialog array. */
	YSRESULT RemoveDialog(class FsGuiDialog *dlg);

	/*! This function removes all modeless dialogs. */
	void RemoveDialogAll(void);

	/*! This function arranges the dialogs according to the arrangement preference
	    specified in each dialog. */
	void ArrangeDialog(void);

	/*! This function kills all focus in the dialogs. */
	void KillDialogFocusAll(void);

	/*! This function returns the number of top-level dialogs that belongs to this canvas. */
	int GetNumDialog(void) const;

	/*! This function returns the top-level dialog of the given index. */
	class FsGuiDialog *GetDialog(int index) const;

	/*! This function returns YSTRUE if the dialog "dlg" is used by this canvas as a modal or modeless dialog. */
	YSBOOL IsDialogUsed(const FsGuiDialog *dlg) const;

	/*! This function returns the active modal dialog.  
	    If it doesn't exist, it returns NULL. */
	class FsGuiDialog *GetActiveModalDialog(void) const;

	/*! This function returns the active modeless dialog.  
	    If it doesn't exist, it returns NULL. */
	class FsGuiDialog *GetActiveDialog(void) const;

	/*! This function sets the active modeless dialog.  If newActiveDlg does not belong 
	    to this canvas, this function returns YSERR.  If  newActiveDlg is NULL,
	    no dialog in this canvas will be active. */
	YSRESULT SetActiveDialog(const class FsGuiDialog *newActiveDlg);

	/*! Control structure (typically a main loop or an interval function) 
	    must call this function to allow this object to non-user and programatic events. */
	void Interval(void);

	/*! Control structure (typically a main loop or an interval function) 
	    must use this function to regularly report the mouse state to the canvas. */
	YSRESULT SetMouseState(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

private:
	// On menu selection, mouse button state must be forgotten by the dialogs so that a button will not be
	// accidentally pressed.
	void ForgetPreviousMouseButtonState(void);

public:
	/*! Control structure (typically a main loop or an interval function) 
	    must use this function to report a key stroke when a key is pressed. */
	YSRESULT KeyIn(int fsKey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt);

	/*! This function updates the window size.  Does nothing if the size stays the same. */
	YSRESULT SetWindowSize(int newWid,int newHei,YSBOOL autoArrangeDialog);

	/*! Control structure (typically a main loop or an interval function) 
	    must use this function to report a character is generated by a key 
	    stroke. */
	YSRESULT CharIn(int c);

	/*! This function draws all the menus and dialogs on the canvas. */
	void Show(void) const;

	/*! This function returns YSTRUE if the canvas needs to be re-drawn. */
	virtual YSBOOL NeedRedraw(void) const;

	/*! This function sets need-redraw flag. */
	virtual void SetNeedRedraw(YSBOOL needRedraw);

	/*! Override this function to write a behavior for key stroke.  
	    Must return YSOK if this function processed the key stroke and the key event must not be sent to other objects.  
	    Return YSERR otherwise. */
	virtual YSRESULT OnKeyDown(int fskey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt);

	/*! Override this function to write a behavior for when a key stroke generated an ASCII character. 
	    Must return YSOK if this function processed the character and the character must not be sent to other objects.
	    Return YSERR otherwise. */
	virtual YSRESULT OnKeyChar(int c);

	/*! Override this function to write a behavior for left button down. 
	    Must return YSOK if the event is processed in this function and must not be sent to other objects. 
	    Return YSERR otherwise.  */
	virtual YSRESULT OnLButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	/*! Override this function to write a behavior for middle button down.
	    Must return YSOK if the event is processed in this function and must not be sent to other objects. 
	    Return YSERR otherwise.  */
	virtual YSRESULT OnMButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	/*! Override this function to write a behavior for right button down. 
	    Must return YSOK if the event is processed in this function and must not be sent to other objects. 
	    Return YSERR otherwise.  */
	virtual YSRESULT OnRButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	/*! Override this function to write a behavior for mouse move. 
	    Must return YSOK if the event is processed in this function and must not be sent to other objects. 
	    Return YSERR otherwise.  */
	virtual YSRESULT OnMouseMove(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my,int prevMx,int prevMy);

	/*! Override this function to write a behavior for left button up. 
	    Must return YSOK if the event is processed in this function and must not be sent to other objects. 
	    Return YSERR otherwise.  */
	virtual YSRESULT OnLButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	/*! Override this function to write a behavior for middle button up. 
	    Must return YSOK if the event is processed in this function and must not be sent to other objects. 
	    Return YSERR otherwise.  */
	virtual YSRESULT OnMButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);

	/*! Override this function to write a behavior for right button up. 
	    Must return YSOK if the event is processed in this function and must not be sent to other objects. 
	    Return YSERR otherwise.  */
	virtual YSRESULT OnRButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
};

////////////////////////////////////////////////////////////

/* } */
#endif
