/* ////////////////////////////////////////////////////////////

  File Name: ysshelldnmtemplate.h
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#ifndef YSSHELLDNMTEMPLATE_IS_INCLUDED
#define YSSHELLDNMTEMPLATE_IS_INCLUDED
/* { */

////////////////////////////////////////////////////////////

template <class SHLCLASS>
class YsShellDnmContainer
{
public:
	class NodeState
	{
	private:
		enum
		{
			FLAG_SHOW=1
		};
		unsigned int flags;
	public:
		YsVec3 relPos;
		YsAtt3 relAtt;
		YsMatrix4x4 tfmCache;

	public:
		NodeState();
		void Initialize(void);
		void SetShow(YSBOOL showFlag);
		const YSBOOL GetShow(void) const;
	};

	class Node : public SHLCLASS
	{
	public:
		YSHASHKEY searchKey;  // Unique within a DNM object.

		YsString nodeName;
		int dnmClassType;
		YsVec3 posInParent;
		YsAtt3 attInParent;
		YsVec3 localRotationCenter,localRotationAxis;

		Node *parent;
		YsArray <Node *> children;

		mutable YsMatrix4x4 currentTfm;
		YsArray <NodeState> stateArray;

		YsArray <YsString> childNameTag; // Used during read/write DNM

		Node();
		void Initialize(void);
	};

	class DnmState
	{
	private:
		mutable YsArray <NodeState> stateArray;
		mutable YsHashTable <YSSIZE_T> nodeKeyToStateArrayIndex;
	public:
		DnmState();
		DnmState(DnmState &&incoming);
		DnmState(const DnmState &incoming);
		DnmState &operator=(const DnmState &incoming);
		DnmState &operator=(DnmState &&incoming);

		void CleanUp(void);
		NodeState &GetState(const Node *node);
		const NodeState &GetState(const Node *node) const;
		void CacheTransformation(const Node *node) const;
		const YsMatrix4x4 &GetTransformation(const Node *node) const;
	};

private:
	class PackedTextFile : public YsTextInputStream
	{
	private:
		YSSIZE_T filePointer;
		YsArray <YsString> dataFile;
	public:
		YsString fn;

		PackedTextFile()
		{
			Initialize();
		}
		void Initialize(void)
		{
			filePointer=0;
			dataFile.CleanUp();
		}
		void AddString(YsString &&incoming)
		{
			dataFile.Increment();
			dataFile.Last().MoveFrom((YsString &)incoming); // Why g++ requires this cast?  Isn't it obvious?
		}
		void AddString(const YsString &incoming)
		{
			dataFile.Increment();
			dataFile.Last()=incoming;
		}
		void BeginRead(void)
		{
			filePointer=0;
		}
		virtual const char *Gets(YsString &str)
		{
			if(filePointer<dataFile.GetN())
			{
				str=dataFile[filePointer];
				++filePointer;
				return str;
			}
			return NULL;
		}
	};
	class ReadDnmVariable
	{
	public:
		enum {
			READSTATE_ERROR=-1,
			READSTATE_NORMAL=0,
			READSTATE_VERTEXARRAY=1,
			READSTATE_PACKEDFILE=99,
			READSTATE_ENDOFFILE=999
		};

		Node *allParent;
		int readState; // -1:Error!!  0:Normal  1:Reading PLG  99:Loading Packed File  999:End of file
		int dnmVersion; // 0: Old dnm  1: Version 1
		int nPackedFileRemain,nVtxRemain;
		YsString readPath;
		int currentNodeType;
		Node *currentNode;
		PackedTextFile *currentPack;

		YsArray <Node *> newNodeArray;

		YsSegmentedArray <PackedTextFile,4> packedFileList;
	};

	static const char *const keyWordSource[];
	static class YsKeyWordList keyWordList;
	YSHASHKEY searchKeySeed;

protected:
	/*! Pointer to the root node. */
	YsArray <Node *> rootArray;

public:
	/*! Default constructor. */
	YsShellDnmContainer();

	/*! Default destructor. */
	~YsShellDnmContainer();

	/*! Returns the bounding box. */
	void GetBoundingBox(YsVec3 bbx[2]) const;

	/*! Returns the bounding box. */
	void GetBoundingBox(YsVec3 &min,YsVec3 &max) const;

	/*! This function cleans up all contents of this object. */
	void CleanUp(void);
private:
	void DeleteShellRecursive(Node *ptr);

public:
	/*! This function creates a shell node. */
	Node *CreateShell(Node *parentPtr);

	/*! This function deletes a shell node. */
	YSRESULT DeleteShell(Node *ptr);

	/*! This function returns the number of children of the node 'ptr'.  If ptr==NULL, it returns the number of root nodes. */
	int GetNumChildNode(Node *ptr) const;

	/*! This function returns the first root node pointer. */
	Node *GetFirstRootNode(void) const;

	/*! This function returns an array of root node pointers. */
	template <const int N>
	void GetRootNodeArray(YsArray <Node *,N> &rootArray) const;

	/*! This function returns an array of root node pointers. */
	YsArray <Node *> GetRootNodeArray(void) const;

	/*! This function returns a constant pointer to all shell nodes in this DNM container. */
	template <const int N>
	void GetNodePointerAll(YsArray <const Node *,N> &allNode) const;

	/*! This function returns a non-constant pointer to all shell nodes in this DNM container. */
	template <const int N>
	void GetNodePointerAll(YsArray <Node *,N> &allNode);

	/*! This function returns a non-constant pointer to all shell nodes in this DNM container. */
	YsArray <Node *> GetNodePointerAll(void);

	/*! This function returns an empty DNM state. */
	DnmState EmptyDnmState(void) const;

	/*! This function clears the state. */
	void ClearState(DnmState &dnmState);

	/*! This function sets up DNM state to 'stateId'. */
	void SetState(DnmState &dnmState,int stateId);
private:
	void SetStateRecursive(DnmState &dnmState,Node *node,int stateId);

public:
	/*! This function caches the transformation in each node. */
	void CacheTransformation(DnmState &dnmState) const;
private:
	void CacheTransformationRecursive(DnmState &dnmState,const Node *node) const;

public:
	/*! This function loads a .DNM file. */
	YSRESULT LoadDnm(YsTextInputStream &inStream,Node *allParent,const char baseFileName[]=NULL);
private:
	YSRESULT BeginReadDnm(ReadDnmVariable &var,Node *allParent,const char fn[]);
	YSRESULT ReadDnmOneLine(ReadDnmVariable &var,const YsString &str);
	YSRESULT EndReadDnm(ReadDnmVariable &var);
	void DeleteChildNodeNameTag(void);
	YSRESULT MakeConnectionAndRootNodeListByChildNameTag(ReadDnmVariable &var);

public:
	/*! This function writes a .DNM file. */
	YSRESULT SaveDnm(YsTextOutputStream &outStream);
};

////////////////////////////////////////////////////////////

template <class SHLCLASS>
YsShellDnmContainer <SHLCLASS>::NodeState::NodeState()
{
	Initialize();
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::NodeState::Initialize(void)
{
	flags=FLAG_SHOW;
	relPos=YsOrigin();
	relAtt=YsZeroAtt();
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::NodeState::SetShow(YSBOOL showFlag)
{
	if(YSTRUE==showFlag)
	{
		flags|=(unsigned int)FLAG_SHOW;
	}
	else
	{
		flags&=(~(unsigned int)FLAG_SHOW);
	}
}

template <class SHLCLASS>
const YSBOOL YsShellDnmContainer <SHLCLASS>::NodeState::GetShow(void) const
{
	if(0!=(flags&FLAG_SHOW))
	{
		return YSTRUE;
	}
	return YSFALSE;
}

////////////////////////////////////////////////////////////

template <class SHLCLASS>
YsShellDnmContainer <SHLCLASS>::Node::Node()
{
	Initialize();
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::Node::Initialize(void)
{
	nodeName="";
	dnmClassType=0;

	posInParent=YsOrigin();
	attInParent=YsZeroAtt();
	localRotationCenter=YsOrigin();
	localRotationAxis=YsXVec();

	parent=NULL;
	children.CleanUp();

	stateArray.CleanUp();
}

////////////////////////////////////////////////////////////

template <class SHLCLASS>
YsShellDnmContainer <SHLCLASS>::DnmState::DnmState()
{
}

template <class SHLCLASS>
YsShellDnmContainer <SHLCLASS>::DnmState::DnmState(const DnmState &incoming)
{
	stateArray.MoveFrom=incoming.stateArray;
	nodeKeyToStateArrayIndex=incoming.nodeKeyToStateArrayIndex;
}

template <class SHLCLASS>
YsShellDnmContainer <SHLCLASS>::DnmState::DnmState(DnmState &&incoming)
{
	stateArray.MoveFrom(incoming.stateArray);
	nodeKeyToStateArrayIndex.MoveFrom(incoming.nodeKeyToStateArrayIndex);
}

template <class SHLCLASS>
typename YsShellDnmContainer <SHLCLASS>::DnmState &YsShellDnmContainer <SHLCLASS>::DnmState::operator=(const DnmState &incoming)
{
	stateArray.MoveFrom=incoming.stateArray;
	nodeKeyToStateArrayIndex=incoming.nodeKeyToStateArrayIndex;
	return *this;
}

template <class SHLCLASS>
typename YsShellDnmContainer <SHLCLASS>::DnmState &YsShellDnmContainer <SHLCLASS>::DnmState::operator=(DnmState &&incoming)
{
	stateArray.MoveFrom(incoming.stateArray);
	nodeKeyToStateArrayIndex.MoveFrom(incoming.nodeKeyToStateArrayIndex);
	return *this;
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::DnmState::CleanUp(void)
{
	stateArray.CleanUp();
	nodeKeyToStateArrayIndex.CleanUp();
}

template <class SHLCLASS>
typename YsShellDnmContainer <SHLCLASS>::NodeState &YsShellDnmContainer <SHLCLASS>::DnmState::GetState(const Node *node)
{
	YSSIZE_T idx;
	if(YSOK!=nodeKeyToStateArrayIndex.FindElement(idx,node->searchKey))
	{
		idx=stateArray.GetN();
		stateArray.Increment();
		stateArray.Last().Initialize();
		nodeKeyToStateArrayIndex.AddElement(node->searchKey,idx);
	}
	return stateArray[idx];
}

template <class SHLCLASS>
const typename YsShellDnmContainer <SHLCLASS>::NodeState &YsShellDnmContainer <SHLCLASS>::DnmState::GetState(const Node *node) const
{
	YSSIZE_T idx;
	if(YSOK!=nodeKeyToStateArrayIndex.FindElement(idx,node->searchKey))
	{
		idx=stateArray.GetN();
		stateArray.Increment();
		stateArray.Last().Initialize();
		nodeKeyToStateArrayIndex.AddElement(node->searchKey,idx);
	}
	return stateArray[idx];
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::DnmState::CacheTransformation(const Node *node) const
{
	YSSIZE_T idx;
	if(YSOK==nodeKeyToStateArrayIndex.FindElement(idx,node->searchKey))
	{
		auto &state=stateArray[idx];

		state.tfmCache.LoadIdentity();
		state.tfmCache.Translate(node->posInParent);

		state.tfmCache.RotateXZ(node->attInParent.h());
		state.tfmCache.RotateZY(node->attInParent.p());
		state.tfmCache.RotateXY(node->attInParent.b());

		state.tfmCache.Translate(state.relPos);

		state.tfmCache.RotateXZ(state.relAtt.h());
		state.tfmCache.RotateZY(state.relAtt.p());
		state.tfmCache.RotateXY(state.relAtt.b());

		state.tfmCache.Translate(-node->localRotationCenter);
	}
}

template <class SHLCLASS>
const YsMatrix4x4 &YsShellDnmContainer <SHLCLASS>::DnmState::GetTransformation(const Node *node) const
{
	YSSIZE_T idx;
	if(YSOK==nodeKeyToStateArrayIndex.FindElement(idx,node->searchKey))
	{
		return stateArray[idx].tfmCache;
	}
	return YsIdentity4x4();
}

////////////////////////////////////////////////////////////

template <class SHLCLASS>
YsShellDnmContainer <SHLCLASS>::YsShellDnmContainer()
{
	searchKeySeed=1;
}

template <class SHLCLASS>
YsShellDnmContainer <SHLCLASS>::~YsShellDnmContainer()
{
	CleanUp();
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::GetBoundingBox(YsVec3 bbx[2]) const
{
	return GetBoundingBox(bbx[0],bbx[1]);
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::GetBoundingBox(YsVec3 &min,YsVec3 &max) const
{
	YsArray <const Node *> allNode;
	GetNodePointerAll(allNode);

	YsBoundingBoxMaker3 mkBbx;
	for(auto &node : allNode)
	{
		YsVec3 shlBbx[2];
		node->GetBoundingBox(shlBbx);
		mkBbx.Add(shlBbx[0]);
		mkBbx.Add(shlBbx[1]);
	}

	mkBbx.Get(min,max);
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::CleanUp(void)
{
	for(auto rootPtr : rootArray)
	{
		DeleteShellRecursive(rootPtr);
	}
	rootArray.CleanUp();
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::DeleteShellRecursive(
    typename YsShellDnmContainer <SHLCLASS>::Node *ptr)
{
	if(NULL!=ptr)
	{
		for(YSSIZE_T childIdx=0; childIdx<ptr->children.GetN(); ++childIdx)
		{
			DeleteShellRecursive(ptr->children[childIdx]);
		}
		ptr->children.CleanUp();
		delete ptr;
	}
}

template <class SHLCLASS>
int YsShellDnmContainer <SHLCLASS>::GetNumChildNode(Node *ptr) const
{
	if(NULL!=ptr)
	{
		return ptr->children.GetN();
	}
	else
	{
		return rootArray.GetN();
	}
}

template <class SHLCLASS>
typename YsShellDnmContainer <SHLCLASS>::Node *YsShellDnmContainer <SHLCLASS>::GetFirstRootNode(void) const
{
	if(0<rootArray.GetN())
	{
		return rootArray[0];
	}
	return NULL;
}

template <class SHLCLASS>
template <const int N>
void YsShellDnmContainer <SHLCLASS>::GetRootNodeArray(YsArray <Node *,N> &rootArray) const
{
	rootArray=this->rootArray;
}

template <class SHLCLASS>
YsArray <typename YsShellDnmContainer <SHLCLASS>::Node *> YsShellDnmContainer <SHLCLASS>::GetRootNodeArray(void) const
{
	return rootArray;
}

template <class SHLCLASS>
typename YsShellDnmContainer <SHLCLASS>::Node *YsShellDnmContainer <SHLCLASS>::CreateShell(
    typename YsShellDnmContainer <SHLCLASS>::Node *parentPtr)
{
	Node *newShl=new Node;
	newShl->searchKey=searchKeySeed++;

	if(NULL!=parentPtr)
	{
		// Verify parentPtr belongs to this DNM.
		Node *ptr=parentPtr;
		while(NULL!=ptr->parent)
		{
			ptr=ptr->parent;
		}

		if(YSTRUE!=rootArray.IsIncluded(ptr))
		{
			YsPrintf("%s %d\n",__FUNCTION__,__LINE__);
			YsPrintf("  Given parentPtr does not belong to this DNM.\n");
			return NULL;
		}

		parentPtr->children.Append(newShl);
		newShl->parent=parentPtr;
	}
	else 
	{
		rootArray.Append(newShl);
		newShl->parent=NULL;
	}

	return newShl;
}

template <class SHLCLASS>
YSRESULT YsShellDnmContainer <SHLCLASS>::DeleteShell(typename YsShellDnmContainer <SHLCLASS>::Node *ptr)
{
	if(NULL!=ptr)
	{
		auto parentPtr=ptr->parent;
		if(NULL!=parentPtr)
		{
			for(YSSIZE_T idx=parentPtr->children.GetN()-1; 0<=idx; --idx)
			{
				if(parentPtr->children[idx]==ptr)
				{
					parentPtr->children.Delete(idx);
				}
			}
		}
		else
		{
			for(YSSIZE_T idx=rootArray.GetN()-1; 0<=idx; --idx)
			{
				if(rootArray[idx]==ptr)
				{
					rootArray.Delete(idx);
				}
			}
		}

		DeleteShellRecursive(ptr);
		return YSOK;
	}
	return YSERR;
}

template <class SHLCLASS>
template <const int N>
void YsShellDnmContainer <SHLCLASS>::GetNodePointerAll(
    YsArray <const typename YsShellDnmContainer <SHLCLASS>::Node *,N> &allNode) const
{
	allNode.CleanUp();

	for(auto rootPtr : rootArray)
	{
		auto i0=allNode.GetN();
		allNode.Append(rootPtr);
		for(YSSIZE_T todoIdx=i0; todoIdx<allNode.GetN(); ++todoIdx)
		{
			const Node *node=allNode[todoIdx];
			for(YSSIZE_T childIdx=0; childIdx<node->children.GetN(); ++childIdx)
			{
				allNode.Append(node->children[childIdx]);
			}
		}
	}
}

template <class SHLCLASS>
template <const int N>
void YsShellDnmContainer <SHLCLASS>::GetNodePointerAll(
    YsArray <typename YsShellDnmContainer <SHLCLASS>::Node *,N> &allNode)
{
	allNode.CleanUp();

	for(auto rootPtr : rootArray)
	{
		auto i0=allNode.GetN();
		allNode.Append(rootPtr);
		for(YSSIZE_T todoIdx=i0; todoIdx<allNode.GetN(); ++todoIdx)
		{
			Node *node=allNode[todoIdx];
			for(YSSIZE_T childIdx=0; childIdx<node->children.GetN(); ++childIdx)
			{
				allNode.Append(node->children[childIdx]);
			}
		}
	}
}

template <class SHLCLASS>
YsArray <typename YsShellDnmContainer <SHLCLASS>::Node *> YsShellDnmContainer <SHLCLASS>::GetNodePointerAll(void)
{
	YsArray <Node *> allNode;
	GetNodePointerAll(allNode);
	return allNode;
}

template <class SHLCLASS>
typename YsShellDnmContainer <SHLCLASS>::DnmState YsShellDnmContainer <SHLCLASS>::EmptyDnmState(void) const
{
	DnmState state;
	return state;
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::ClearState(DnmState &dnmState)
{
	dnmState.Initialize();
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::SetState(DnmState &dnmState,int stateId)
{
	for(auto node : rootArray)
	{
		SetStateRecursive(dnmState,node,stateId);
	}
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::SetStateRecursive(DnmState &dnmState,Node *node,int stateId)
{
	auto &state=dnmState.GetState(node);
	if(YSTRUE==node->stateArray.IsInRange(stateId))
	{
		state=node->stateArray[stateId];
	}
	for(auto childNode : node->children)
	{
		SetStateRecursive(dnmState,childNode,stateId);
	}
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::CacheTransformation(DnmState &dnmState) const
{
	for(auto node : rootArray)
	{
		CacheTransformationRecursive(dnmState,node);
	}
}

template <class SHLCLASS>
void YsShellDnmContainer <SHLCLASS>::CacheTransformationRecursive(DnmState &dnmState,const Node *node) const
{
	dnmState.CacheTransformation(node);
	for(auto childNode : node->children)
	{
		CacheTransformationRecursive(dnmState,childNode);
	}
}

template <class SHLCLASS>
YSRESULT YsShellDnmContainer<SHLCLASS>::LoadDnm(YsTextInputStream &inStream,Node *allParent,const char dnmFileName[])
{
	YsObjectFromHeap <ReadDnmVariable> var;

	YsString idStr;
	int line=1;
	if(NULL!=inStream.Gets(idStr))
	{
		++line;

		idStr.Capitalize();
		if(strncmp(idStr,"DYNAMODEL",9)==0)
		{
			if(BeginReadDnm(var,allParent,dnmFileName)==YSOK)
			{
				YsString buf;
				while(NULL!=inStream.Gets(buf))
				{
					line++;
					if(ReadDnmOneLine(var,buf)!=YSOK)
					{
						printf("Line %d\n",line);
						EndReadDnm(var);
						goto ERREND;
					}
				}
				if(EndReadDnm(var)==YSOK)
				{
					return YSOK;
				}
			}
		}
	}

ERREND:
	return YSERR;
}

template <class SHLCLASS>
YSRESULT YsShellDnmContainer<SHLCLASS>::BeginReadDnm(ReadDnmVariable &var,Node *allParent,const char fn[])
{
	var.allParent=allParent;
	var.readState=0;
	var.currentNode=NULL;
	var.nPackedFileRemain=0;
	var.dnmVersion=0;

	if(NULL!=fn)
	{
		YsString ful(fn),fil;
		ful.SeparatePathFile(var.readPath,fil);
	}
	else
	{
		var.readPath.Set(".");
	}
	return YSOK;
}

template <class SHLCLASS>
const char *const YsShellDnmContainer<SHLCLASS>::keyWordSource[]=
{
	"DYNAMODEL", // 0
	"DNMVER",    // 1
	"PCK",       // 2
	"SRF",       // 3
	"PLG",       // 4
	"LIN",       // 5
	"CLA",       // 6
	"NST",       // 7
	"STA",       // 8
	"NCH",       // 9
	"REL",       // 10
	"CLD",       // 11
	"FIL",       // 12
	"VTX",       // 13
	"NVT",       // 14
	"COL",       // 15
	"POS",       // 16
	"CNT",       // 17
	"END",       // 18
	"PAX",       // 19
	NULL
};

template <class SHLCLASS>
YsKeyWordList YsShellDnmContainer<SHLCLASS>::keyWordList;

template <class SHLCLASS>
YSRESULT YsShellDnmContainer<SHLCLASS>::ReadDnmOneLine(ReadDnmVariable &var,const YsString &str)
{
	YsArray <YsString,16> av;
	YsString buf=str;

	if(keyWordList.GetN()==0)
	{
		keyWordList.MakeList(keyWordSource);
	}

	buf.Capitalize();

	if(var.readState==ReadDnmVariable::READSTATE_PACKEDFILE/*99*/)
	{
		var.currentPack->AddString(str);
		var.nPackedFileRemain--;
		if(var.nPackedFileRemain==0)
		{
			var.readState=ReadDnmVariable::READSTATE_NORMAL/*0*/;
		}
		return YSOK;
	}
	else if(YSOK==buf.Arguments(av) && 0<av.GetN())
	{
		auto ac=av.GetN();
		if(var.readState==ReadDnmVariable::READSTATE_NORMAL/*0*/)
		{
			if(av[0][0]=='#')    // 2003/03/25
			{                         // 2003/03/25
				// Comment line       // 2003/03/25
				return YSOK;          // 2003/03/25
			}                         // 2003/03/25

			const int cmdId=keyWordList.GetId(av[0]);
			switch(cmdId)
			{
			case 0:   // "DYNAMODEL", // 0
				// Ignore
				return YSOK;
			case 1:   // "DNMVER",    // 1
				if(ac>=2)
				{
					var.dnmVersion=atoi(av[1]);
					return YSOK;
				}
				break;
			case 2:   // "PCK",       // 2
				if(ac>=3)
				{
					var.packedFileList.Increment();
					var.packedFileList.Last().Initialize();
					var.packedFileList.Last().fn=av[1];
					var.currentPack=&var.packedFileList.Last();
					var.nPackedFileRemain=atoi(av[2]);
					if(var.nPackedFileRemain>0)
					{
						var.readState=ReadDnmVariable::READSTATE_PACKEDFILE/*99*/;
					}
					return YSOK;
				}
				break;
			case 3:   // "SRF",       // 3
				if(ac>=2)
				{
					auto newNode=CreateShell(NULL);
					var.newNodeArray.Append(newNode);
					var.currentNode=newNode;
					var.currentNode->nodeName=av[1];
					var.currentNodeType=0;
					return YSOK;
				}
				break;
			case 4:   // "PLG",       // 4
				if(ac>=2)
				{
					auto newNode=CreateShell(NULL);
					var.newNodeArray.Append(newNode);
					var.currentNode=newNode;
					var.currentNode->nodeName=av[1];
					var.currentNodeType=1;
					return YSOK;
				}
				break;
			case 5:   // "LIN",       // 5
				if(ac>=2)
				{
					auto newNode=CreateShell(NULL);
					var.newNodeArray.Append(newNode);
					var.currentNode=newNode;
					var.currentNode->nodeName=av[1];
					var.currentNodeType=2;
					return YSOK;
				}
				break;
			case 6:   // "CLA",       // 6
				if(ac>=2 && var.currentNode!=NULL)
				{
					var.currentNode->dnmClassType=atoi(av[1]);
					return YSOK;
				}
				break;
			case 7:   // "NST",       // 7
				// Ignore
				return YSOK;
			case 8:   // "STA",       // 8
				if(ac>=8 && var.currentNode!=NULL)
				{
					NodeState sta;
					sta.relPos.Set(atof(av[1]),atof(av[2]),atof(av[3]));
					sta.relAtt.Set(atof(av[4])*YsPi/32768,atof(av[5])*YsPi/32768,atof(av[6])*YsPi/32768);
					sta.SetShow(atoi(av[7])==0 ? YSFALSE : YSTRUE);
					var.currentNode->stateArray.Append(sta);
					return YSOK;
				}
				break;
			case 9:   // "NCH",       // 9
				// Nothing to do
				return YSOK;
			case 10:  // "REL",       // 10
				// Only "DEP" is accepted, thus, ignored
				return YSOK;
			case 11:  // "CLD",       // 11
				if(var.currentNode!=NULL && ac>=2)
				{
					var.currentNode->childNameTag.Increment();
					var.currentNode->childNameTag.Last()=av[1];
					return YSOK;
				}
				break;
			case 12:  // "FIL",       // 12
				if(var.currentNode!=NULL && ac>=2)
				{
					if(var.currentNodeType==0)
					{
						YSBOOL isPacked=YSFALSE;
						for(auto &pack : var.packedFileList)
						{
							if(strcmp(pack.fn,av[1])==0)
							{
								pack.BeginRead();
								var.currentNode->LoadSrf(pack);
								isPacked=YSTRUE;
								break;
							}
						}
						if(isPacked!=YSTRUE)
						{
							YsString ful;
							ful.MakeFullPathName(var.readPath,av[1]);

							FILE *fp=fopen(ful,"r");
							if(NULL!=fp)
							{
								YsTextFileInputStream srfInStream(fp);
								if(var.currentNode->LoadSrf(srfInStream)!=YSOK)
								{
									YsPrintf("Cannot read %s\n",ful.Txt());
									var.readState=ReadDnmVariable::READSTATE_ERROR/*-1*/;
									fclose(fp);
									return YSERR;
								}
								fclose(fp);
							}
						}
					}
					return YSOK;
				}
				break;
			case 13:  // "VTX",       // 13
				if(var.currentNode!=NULL && ac==7)
				{
					if(var.currentNodeType==2)
					{
						double l;
						YsVec3 p1,p2,u,v,w,vv,ww;
						p1.Set(atof(av[1]),atof(av[2]),atof(av[3]));
						p2.Set(atof(av[4]),atof(av[5]),atof(av[6]));
						u=p2-p1;
						u.Normalize();
						v=u.GetArbitraryPerpendicularVector();
						v.Normalize();
						w=u^v;
						l=(p2-p1).GetLength();
						vv=v*0.015;  // vv=v*l/50.0;  2003/03/26
						ww=w*0.015;  // ww=w*l/50.0;  2003/03/26

						YsShellVertexHandle vtHd[8],quad[4];
						YsShellPolygonHandle plHd;
						vtHd[0]=var.currentNode->AddVertex(p1+vv+ww);
						vtHd[1]=var.currentNode->AddVertex(p1+vv-ww);
						vtHd[2]=var.currentNode->AddVertex(p1-vv-ww);
						vtHd[3]=var.currentNode->AddVertex(p1-vv+ww);
						vtHd[4]=var.currentNode->AddVertex(p2+vv+ww);
						vtHd[5]=var.currentNode->AddVertex(p2+vv-ww);
						vtHd[6]=var.currentNode->AddVertex(p2-vv-ww);
						vtHd[7]=var.currentNode->AddVertex(p2-vv+ww);

						quad[0]=vtHd[0];
						quad[1]=vtHd[1];
						quad[2]=vtHd[2];
						quad[3]=vtHd[3];
						plHd=var.currentNode->AddPolygon(4,quad);
						var.currentNode->SetPolygonColor(plHd,YsWhite());
						var.currentNode->SetPolygonNormal(plHd,-u);

						quad[0]=vtHd[7];
						quad[1]=vtHd[6];
						quad[2]=vtHd[5];
						quad[3]=vtHd[4];
						plHd=var.currentNode->AddPolygon(4,quad);
						var.currentNode->SetPolygonColor(plHd,YsWhite());
						var.currentNode->SetPolygonNormal(plHd,u);

						quad[0]=vtHd[1];
						quad[1]=vtHd[0];
						quad[2]=vtHd[4];
						quad[3]=vtHd[5];
						plHd=var.currentNode->AddPolygon(4,quad);
						var.currentNode->SetPolygonColor(plHd,YsWhite());
						var.currentNode->SetPolygonNormal(plHd,v);

						quad[0]=vtHd[2];
						quad[1]=vtHd[1];
						quad[2]=vtHd[5];
						quad[3]=vtHd[6];
						plHd=var.currentNode->AddPolygon(4,quad);
						var.currentNode->SetPolygonColor(plHd,YsWhite());
						var.currentNode->SetPolygonNormal(plHd,-w);

						quad[0]=vtHd[3];
						quad[1]=vtHd[2];
						quad[2]=vtHd[6];
						quad[3]=vtHd[7];
						plHd=var.currentNode->AddPolygon(4,quad);
						var.currentNode->SetPolygonColor(plHd,YsWhite());
						var.currentNode->SetPolygonNormal(plHd,-v);

						quad[0]=vtHd[0];
						quad[1]=vtHd[3];
						quad[2]=vtHd[7];
						quad[3]=vtHd[4];
						plHd=var.currentNode->AddPolygon(4,quad);
						var.currentNode->SetPolygonColor(plHd,YsWhite());
						var.currentNode->SetPolygonNormal(plHd,w);

						// var.currentNode->RecomputeNormalAndFlipDirection();
						// var.currentNode->SortPolygonByColor();
						// var.currentNode->PrecomputePolygCenterAfterSortingByColor();
					}
					return YSOK;
				}
				break;
			case 14:  //  "NVT",       // 14
				if(var.currentNode!=NULL && ac>=2)
				{
					var.nVtxRemain=atoi(av[1]);
					if(var.nVtxRemain>0)
					{
						var.readState=ReadDnmVariable::READSTATE_VERTEXARRAY/*1*/;
					}
					return YSOK;
				}
				break;
			case 15:  // "COL",       // 15
				if(var.currentNode!=NULL && ac>=2)
				{
					YsColor col;
					if(ac>=4)
					{
						col.SetIntRGB(atoi(av[1]),atoi(av[2]),atoi(av[3]));
					}
					else
					{
						col.Set15BitRGB(atoi(av[1]));
					}
					for(auto plHd : var.currentNode->AllPolygon())
					{
						var.currentNode->SetPolygonColor(plHd,col);
					}
					return YSOK;
				}
				break;
			case 16:  // "POS",       // 16
				if(var.currentNode!=NULL && ac>=7)
				{
					double x,y,z,h,p,b;
					x=atof(av[1]);
					y=atof(av[2]);
					z=atof(av[3]);
					h=atof(av[4])*YsPi/32768;
					p=atof(av[5])*YsPi/32768;
					b=atof(av[6])*YsPi/32768;
					var.currentNode->posInParent.Set(x,y,z);
					var.currentNode->attInParent.Set(h,p,b);
					return YSOK;
				}
				break;
			case 17:  // "CNT",       // 17
				if(var.currentNode!=NULL && ac>=4)
				{
					double x,y,z;
					x=atof(av[1]);
					y=atof(av[2]);
					z=atof(av[3]);
					var.currentNode->localRotationCenter.Set(x,y,z);
					return YSOK;
				}
				break;
			case 18:  // "END",       // 18
				if(var.currentNode!=NULL)
				{
					var.currentNode=NULL;
				}
				else
				{
					var.readState=ReadDnmVariable::READSTATE_ENDOFFILE/*999*/;
				}
				return YSOK;
			case 19:  // "PAX",       // 19
				if(var.currentNode!=NULL && ac>=4)
				{
					double x,y,z;
					x=atof(av[1]);
					y=atof(av[2]);
					z=atof(av[3]);
					var.currentNode->localRotationAxis.Set(x,y,z);
					return YSOK;
				}
				break;
			}
		}
		else if(var.readState==ReadDnmVariable::READSTATE_VERTEXARRAY/*1*/)
		{
			YsShellVertexHandle vtHd;
			if(ac>=4)
			{
				YsVec3 pos;
				pos.Set(atof(av[1]),atof(av[2]),atof(av[3]));
				var.currentNode->AddVertex(pos);
				var.nVtxRemain--;
				if(var.nVtxRemain==0)
				{
					vtHd=NULL;
					YsArray <YsShellVertexHandle,4> plVtHd;
					YsArray <YsVec3> plVtPos;
					while(YSOK==var.currentNode->MoveToNextVertex(vtHd))
					{
						plVtHd.Append(vtHd);
						var.currentNode->GetVertexPosition(pos,vtHd);
						plVtPos.Append(pos);
					}

					auto plHd=var.currentNode->AddPolygon(plVtHd);
					var.currentNode->SetPolygonTwoSidedFlag(plHd,YSTRUE);

					YsVec3 nom;
					YsGetAverageNormalVector(nom,plVtPos.GetN(),plVtPos);
					nom.Normalize();
					var.currentNode->SetPolygonNormal(plHd,nom);

					// var.currentNode->RecomputeNormalAndFlipDirection();
					// var.currentNode->SortPolygonByColor();
					// var.currentNode->PrecomputePolygCenterAfterSortingByColor();

					var.readState=ReadDnmVariable::READSTATE_NORMAL/*0*/;
				}
				return YSOK;
			}
			else
			{
				YsPrintf("Error during reading polygon node.\n");
				YsPrintf("  >> %s\n",str.Txt());
				var.readState=ReadDnmVariable::READSTATE_ERROR/*-1*/;
				return YSERR;
			}
		}
		else if(var.readState==ReadDnmVariable::READSTATE_ENDOFFILE/*999*/)
		{
			return YSOK;
		}

		YsPrintf("Error during reading DNM\n");
		YsPrintf("  >> %s\n",str.Txt());
		return YSERR;
	}
	return YSOK;
}

template <class SHLCLASS>
YSRESULT YsShellDnmContainer<SHLCLASS>::EndReadDnm(ReadDnmVariable &var)
{
	MakeConnectionAndRootNodeListByChildNameTag(var);
	DeleteChildNodeNameTag();
	return YSOK;
}

template <class SHLCLASS>
void YsShellDnmContainer<SHLCLASS>::DeleteChildNodeNameTag(void)
{
	YsArray <Node *> allNode;
	GetNodePointerAll(allNode);
	for(auto node : allNode)
	{
		node->childNameTag.ClearDeep();
	}
}

template <class SHLCLASS>
YSRESULT YsShellDnmContainer<SHLCLASS>::MakeConnectionAndRootNodeListByChildNameTag(ReadDnmVariable &var)
{
	// At this point, all nodes are tentatively root nodes.
	auto dnmVersion=var.dnmVersion;
	auto allParent=var.allParent;

	YsArray <Node *> allNode=var.newNodeArray;
	YsKeyStore newNodeKey;
	for(auto node : allNode)
	{
		node->parent=NULL;
		node->children.CleanUp();
		newNodeKey.AddKey(node->searchKey);
	}
	for(YSSIZE_T idx=rootArray.GetN()-1; 0<=idx; --idx)
	{
		if(YSTRUE==newNodeKey.IsIncluded(rootArray[idx]->searchKey))
		{
			rootArray.Delete(idx);
		}
	}

	for(auto parent : allNode)
	{
		for(auto child : allNode)
		{
			for(auto &childName : parent->childNameTag)
			{
				if(0==strcmp(childName,child->nodeName))
				{
					YSBOOL isLoop=YSFALSE;
					Node *loopCheck=parent;
					while(loopCheck!=NULL)
					{
						if(loopCheck==child)
						{
							YsPrintf("Loop in DNM file!\n");
							isLoop=YSTRUE;
							break;
						}
						loopCheck=loopCheck->parent;
					}

					if(isLoop!=YSTRUE)
					{
						child->parent=parent;
						parent->children.Append(child);
					}
					break;
				}
			}
		}
	}

	YsArray <Node *> newRootNode;
	for(auto node : allNode)
	{
		if(node->parent==NULL)
		{
			newRootNode.Append(node);
		}
	}

	if(dnmVersion==0)  
	{
		// Mimicing the bug of the old DNM class.  
		//   Older version of DNM format allowed only one root node.  
		//   All node except the very first root node are automatically made a child of the first root node.
		for(YSSIZE_T i=newRootNode.GetN()-1; i>0; i--)
		{
			newRootNode[0]->children.Append(newRootNode[i]);
			newRootNode[i]->parent=newRootNode[0];
			newRootNode.Delete(i);
		}
	}

	if(NULL==allParent)
	{
		rootArray.Append(newRootNode);
	}
	else
	{
		allParent->children.Append(newRootNode);
	}

	return YSOK;
}

////////////////////////////////////////////////////////////


/* } */
#endif
